"""
Thomas Wick, Sebastian Kinnewig
Leibniz University Hannover
Institute of Applied Mathematics
Germany

https://www.pec3.org/events/school2022/

Exercise Class 1 - Iterative Solver and Gradient Decent Methods

Exercise 3 - PCG
"""


"""
===== PACKAGES =====
"""
# Basic math tools 
import numpy as np

# Library to create plots 
import matplotlib.pyplot as plt 


"""
===== Exercise 3: (a) SOR Preconditioner =====
Implement the SOR Preconditioner as descibed on slide 63
"""

def diagonal( A ):
  # get the size of A
  size = len( A )

  # create a new output matrix
  D = np.zeros((size, size))

  # copy the corresponding entries
  for i in range(size):
    D[i][i] = A[i][i]

  return D

def lower_triangular( A ):
  # get the size of A
  size = len( A )

  # create a new output matrix
  L = np.zeros((size, size))

  # copy the corresponding entries
  for i in range(1, size):
    for j in range(0, i):
      L[i][j] = A[i][j];

  return L;

def upper_triangular( A ):
  # get the size of A
  size = len( A )

  # create a new output matrix
  U = np.zeros((size, size))

  # copy the corresponding entries
  for i in range(size):
    for j in range(i + 1, size):
      U[i][j] = A[i][j];

  return U;


def inverse_diagonal( A ):
  # get the size of A
  size = len( A )

  # create a new output matrix
  D = np.zeros((size, size))

  # copy the corresponding entries
  for i in range(size):
    D[i][i] = 1.0 / A[i][i]

  return D

def ssor( A, w = 1.0):
  # get the size of A
  size = len( A )

  # part 1
  C1 = diagonal(A) + ( w * lower_triangular(A) )

  # inverse diagonal:
  C2 = inverse_diagonal( A )

  # part 3
  C3 = diagonal(A) + ( w * upper_triangular(A) )

  return np.dot(np.dot(C1, C2), C3)



"""
===== Exercise 3: (b) Implement the PCG =====
Implement the PCG method as described on slide 60
"""

def pcg(A, f, P_inverse, epsilon = 1e-5, max_iterations = 10000):
  # get the size of A
  size = len( A )

  # iteration start:
  # create an initial guess x0
  x_old = np.zeros( size )
  r_old = f - np.dot(A, x_old)
  p_old = np.dot(P_inverse, r_old)
  d_old = p_old
    
  # iteration step
  for k in range(max_iterations):

    Adk   = np.dot(A, d_old)
    
    ak    = np.dot(r_old, p_old) / np.dot(Adk, d_old)

    x_new = x_old + (ak * p_old)
    r_new = r_old - (ak * Adk)
    p_new = np.dot(P_inverse, r_new)

    if np.sqrt( np.dot(r_new, r_new) ) < epsilon:
      print("Converged after ", k, " steps")
      x_old = x_new.copy()
      break

    bk    = np.dot(r_new, p_new) / np.dot(r_old, p_old)
    d_new = p_new + (bk * d_old)

    # update
    x_old = x_new.copy()
    r_old = r_new.copy()
    p_old = p_new.copy()
    d_old = d_new.copy()

  # return the result
  return x_old

"""
===== Exercise 3: (c) Comparison =====
Compare different preconditioners

Here we use a random generated matrix for the test case, to ensure that the matrix we 
are considering is symmetric and semi positiv definit we generate first randomly a 
matrix K and than compute A = K.K^T, so A is ensured to be SPD and symmetric.
"""

# create the n x n identity matrix
def identity ( n ):
  result = np.zeros((n,n))

  for i in range(n):
    result[i][i] = 1.0

  return result

n = 100 
K = np.random.randint(10, size=(n, n))
A = np.dot(K, K.T)
f = np.random.randint(10, size=(n))
print("Using the identity matrix as preconditioner:")
x = pcg(A, f, identity(n) )
print("Using SOR as preconditioner")
x = pcg(A, f, np.linalg.inv( ssor(A, 0.5) )) # we need P^(-1) but so far we only computed P, as we now just want to test our pcg method, we are using the invert function provided by numpy
